package net.outlyer.plugins;

/*
 * Copyright (c) 2008, Toni Corvera. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import java.io.IOException;
import java.io.LineNumberReader;
import java.lang.reflect.Field;
import java.net.URI;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.script.Bindings;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;

/**
 * Default implementation of a PluginEnvironment.
 */
final class PluginEnvironmentImpl extends PluginEnvironment {
    
    static class UnsupportedAPIException extends PluginException {
        public UnsupportedAPIException(int providedApi) {
            super("Plugin has unsupported API ("+providedApi+")");
        }
    }
    
    static class UnsupportedTypeException extends PluginException {
        public UnsupportedTypeException(String pluginType) {
            super("Plugin has unsupported type ("+pluginType+")");
        }        
    }
    
    private final Map<String, Object> exportedObjects;
    private BasePluginObject pluginObject;
    private StringBuilder boilerPlate;
    private List<String> acceptedTypes = null;
    private int maxApi=API.REVISION;
    int linesToCheckForSupport = 5;
    
    static final String EXPORTED_SANDBOX_VARIABLE = "$net.outlyer.runtime.sandbox";
    
    {
        exportedObjects = new HashMap();
        pluginObject = null;
        boilerPlate = new StringBuilder();
    }
    
    private void enableSandboxAccess(final String name, final SandboxAccessor sa) {
        try {
            Field f = sa.getClass().getField("_getSandbox");
        }
        catch (final NoSuchFieldException e) {
            throw new IllegalArgumentException("object " + name + " must comply " +
                    "with the contract of SandboxAccessor");
        }

        final String callback = name + "._getSandbox";

        boilerPlate.append(callback).append("=function() {")
                   .append("  return ")
                   .append(EXPORTED_SANDBOX_VARIABLE).append(";\n};");

    }
    
    /**
     * {@inheritDoc}
     */
    public void exportGlobalObject(final String name, final Object object) {
        if (null == name) {
            throw new IllegalArgumentException("Can't export an object with" +
                                               "a null name");
        }
        if (name.equals("plugin")) {
            throw new IllegalArgumentException("Can't export an object named " +
                                        "plugin, use setPluginObject() instead");
        }
        exportedObjects.put(name, object);
        
        if (object instanceof SandboxAccessor) {
            enableSandboxAccess(name, (SandboxAccessor)object);
        }
    }
    
    /**
     * {@inheritDoc}
     */
    public void setPluginObject(BasePluginObject pObject) {
        pluginObject = pObject;
    }
    
    PluginProperties checkForSupport(final URI uri) throws PluginException {
        final ScriptEngine rhino = new ScriptEngineManager().getEngineByName("rhino");

        try {
            // Define a Plugin object with a field named type
            rhino.eval("var plugin={ type: null, apiVersion: null };");
        }
        catch (ScriptException e) {
            assert false;
            throw new IllegalStateException("Unknown error encountered");
        }

        // First of all check the script for support
        Object type = null, name = null;
        Object apiVersion = null;
        // Try to eval each line, since it executes in a sandbox this is
        // *relatively* safe
        String line;
        try {
            final LineNumberReader script = new LineNumberReader(new PluginReader(uri));
            try {
                while ((null == type || null == apiVersion || null == name) &&
                        (null != (line = script.readLine())) &&
                        script.getLineNumber() < linesToCheckForSupport) {
                    try {
                        rhino.eval(line);
                        // Note that undefined doesn't convert to null!
                        if (null == type) {
                            rhino.eval("$_1_$ = (undefined===plugin.type)?null:plugin.type;");
                            type = rhino.get("$_1_$");
                        }
                        if (null == apiVersion) {
                            rhino.eval("$_1_$ = (undefined===plugin.apiVersion)?null:plugin.apiVersion;");
                            apiVersion = rhino.get("$_1_$");
                        }
                        if (null == name) {
                            rhino.eval("$_1_$ = (undefined===plugin.name)?null:plugin.name;");
                            name = rhino.get("$_1_$");
                        }
                    }
                    catch (final ScriptException e) {
                    // Exceptions are to be expected since none of the guaranteed
                    // plugin tools are provided in this context they'll try to
                    // access unexisting objects
                    }
                }
            }
            finally {
                script.close();
            }
        }
        catch (final IOException e) {
            throw new PluginExecutionException("Failed while reading plugin: " + e.getMessage(), e);
        }

        if (null == type) {
            type = "";
        }
        if (null == apiVersion) {
            apiVersion = 0;
        }
        // name can be null
        
        final PluginProperties pp = new PluginProperties();

        try {
            pp.apiVersion = Double.valueOf(apiVersion.toString()).intValue();
            pp.type = type.toString();
            pp.name = (null != name) ? name.toString(): null;
        }
        catch (final NumberFormatException e) {
            throw new PluginException("Incorrect value for apiVersion provided, must be integer");
        }
        
        if (null != acceptedTypes) {
            if (!acceptedTypes.contains(pp.type)) {
                throw new UnsupportedTypeException(pp.type);
            }
        }
        if (pp.apiVersion > maxApi) {
            throw new UnsupportedAPIException(pp.apiVersion);
        }
        
        return pp;
    }

    /**
     * {@inheritDoc}
     */
    public Sandbox createSandbox(final URI forUri) throws PluginException {
        final ScriptEngine rhino = new ScriptEngineManager().getEngineByName("rhino");
        
        final PluginProperties pp = checkForSupport(forUri);

        if (null == pluginObject) {
            pluginObject = new PluginObject();
        }

        try {
            final PluginObject po = (PluginObject) pluginObject.clone();

            final Bindings bindings = rhino.createBindings();

            for (final String objectName : exportedObjects.keySet()) {
                bindings.put(objectName, exportedObjects.get(objectName));
            }
            bindings.put("plugin", po);
            enableSandboxAccess("plugin", po);

            return new SandboxImpl(forUri, bindings, pp, boilerPlate.toString());
        }
        catch (final CloneNotSupportedException e) {
            assert false;
            throw new IllegalStateException("Incorrect implementation");
        }
    }

    /**
     * {@inheritDoc}
     */
    public void setAcceptedTypes(final String ... pluginTypes) {
        if (null != pluginTypes) {
            for (final String type : pluginTypes) {
                if (null == type) {
                    throw new IllegalArgumentException("The set of plugin types" +
                            " must be either null, or a set of non-null values");
                }
            }
        }
        acceptedTypes = Arrays.asList(pluginTypes);
    }
}
